@@ -1,15 +1,23 @@
+from __future__ import print_function
+from __future__ import unicode_literals
+
 import errno
 import glob
-import hexagonit.recipe.download
 import logging
 import os
 import re
 import shutil
+import stat
 import subprocess
-import urllib
 import zc.buildout
 
-from string import strip
+import six.moves.urllib as urllib
+from six import text_type as str
+
+import hexagonit.recipe.download
+
+strip = lambda x:x.strip()  # noqa
+
 
 class Recipe(object):
     """zc.buildout recipe for compiling and installing software"""
@@ -21,9 +29,9 @@ class Recipe(object):
         self.log = logging.getLogger(name)
 
         options['location'] = os.path.join(
-                buildout['buildout']['parts-directory'],
-                self.name,
-            )
+            buildout['buildout']['parts-directory'],
+            self.name,
+        )
 
         if 'gems' not in options:
             self.log.error("Missing 'gems' option.")
@@ -37,24 +45,27 @@ class Recipe(object):
 
     def run(self, cmd, environ=None):
         """Run the given ``cmd`` in a child process."""
-        log = logging.getLogger(self.name)
-
         env = os.environ.copy()
         if environ:
             env.update(environ)
 
         try:
-            retcode = subprocess.call(cmd, shell=True, env=env)
-
-            if retcode < 0:
-                log.error('Command received signal %s: %s' % (-retcode, cmd))
-                raise zc.buildout.UserError('System error')
-            elif retcode > 0:
-                log.error('Command failed with exit code %s: %s' % (retcode, cmd))
-                raise zc.buildout.UserError('System error')
-        except OSError, e:
-            log.error('Command failed: %s: %s' % (e, cmd))
+            subprocess.check_output(cmd, env=env)
+        except OSError as e:
+            self.log.error('Command failed: %s: %s' % (e, cmd))
             raise zc.buildout.UserError('System error')
+        except subprocess.CalledProcessError as e:
+            self.log.error(e.output)
+            if e.returncode < 0:
+                self.log.error('Command received signal %s: %s' % (
+                    -e.returncode, e.cmd
+                ))
+                raise zc.buildout.UserError('System error')
+            elif e.returncode > 0:
+                self.log.error('Command failed with exit code %s: %s' % (
+                    e.returncode, e.cmd
+                ))
+                raise zc.buildout.UserError('System error')
 
     def update(self):
         pass
@@ -65,10 +76,12 @@ class Recipe(object):
     def _get_env_override(self, env):
         env = filter(None, map(strip, env.splitlines()))
         try:
-            env = [map(strip, line.split('=', 1)) for line in env]
-        except ValueError: # Unpacking impossible
+            env = list([(key, val) for key, val in [
+                map(strip, line.split('=', 1)) for line in env
+            ]])
+        except ValueError:  # Unpacking impossible
             self.log.error("Every environment line should contain a '=' sign")
-            zc.buildout.UserError('Configuration error')
+            raise zc.buildout.UserError('Configuration error')
         return env
 
     def _get_env(self):
@@ -80,35 +93,37 @@ class Recipe(object):
         env = {
             'GEM_HOME': '%(PREFIX)s/lib/ruby/gems/1.8' % s,
             'RUBYLIB': self._join_paths(
-                    '%(RUBYLIB)s',
-                    '%(PREFIX)s/lib',
-                    '%(PREFIX)s/lib/ruby',
-                    '%(PREFIX)s/lib/site_ruby/1.8',
-                ) % s,
+                '%(RUBYLIB)s',
+                '%(PREFIX)s/lib',
+                '%(PREFIX)s/lib/ruby',
+                '%(PREFIX)s/lib/site_ruby/1.8',
+            ) % s,
             'PATH': self._join_paths(
-                    '%(PATH)s',
-                    '%(PREFIX)s/bin',
-                ) % s,
+                '%(PATH)s',
+                '%(PREFIX)s/bin',
+            ) % s,
         }
         env_override = self.options.get('environment', '')
         env_override = self._get_env_override(env_override)
-        env.update({k: v % env for k, v in env_override})
+        env.update({k: (v % env) for k, v in env_override})
         return env
 
     def _get_latest_rubygems(self):
         if self.url:
             version = self.version
             if not version:
-                version = re.search(r'rubygems-([0-9.]+).zip$', self.url).group(1)
+                version = (
+                    re.search(r'rubygems-([0-9.]+).zip$', self.url).group(1)
+                )
             return (self.url, version)
 
         if self.version:
             return ('http://production.cf.rubygems.org/rubygems/'
                     'rubygems-%s.zip' % self.version, self.version)
 
-        f = urllib.urlopen('http://rubygems.org/pages/download')
+        f = urllib.request.urlopen('http://rubygems.org/pages/download')
         s = f.read()
-        s = unicode(s)
+        s = str(s)
         f.close()
         r = re.search(r'http://production.cf.rubygems.org/rubygems/'
                       r'rubygems-([0-9.]+).zip', s)
@@ -117,41 +132,50 @@ class Recipe(object):
             version = r.group(1)
             return (url, version)
         else:
-            return None
+            self.log.error("Can't find latest rubygems version.")
+            raise zc.buildout.UserError('Configuration error')
 
     def _install_rubygems(self):
         url, version = self._get_latest_rubygems()
 
-        opt = self.options.copy()
-        opt['url'] = url
-        opt['destination'] = self.buildout['buildout']['parts-directory']
-        hexagonit.recipe.download.Recipe(self.buildout, self.name,
-                                         opt).install()
+        options = {
+            'url': url,
+            'destination': self.buildout['buildout']['parts-directory'],
+        }
+        HexagonitDownload = hexagonit.recipe.download.Recipe
+        recipe = HexagonitDownload(self.buildout, self.name, options)
+        recipe.install()
 
         current_dir = os.getcwd()
         try:
             os.mkdir(self.options['location'])
-        except OSError, e:
+        except OSError as e:
             if e.errno == errno.EEXIST:
                 pass
+            else:
+                self.log.error((
+                    "IO error while creating '%s' directory."
+                ) % self.options['location'])
+                raise zc.buildout.UserError('Configuration error')
 
         srcdir = os.path.join(self.buildout['buildout']['parts-directory'],
                               'rubygems-%s' % version)
         os.chdir(srcdir)
 
-        try:
-            env = self._get_env()
-            env['PREFIX'] = self.options['location']
+        env = self._get_env()
+        env['PREFIX'] = self.options['location']
 
-            s = {
-                'OPTIONS': ' '.join([
-                        '--prefix=%s' % self.options['location'],
-                        '--no-rdoc',
-                        '--no-ri',
-                    ]),
-                'RUBY': self.ruby_executable,
-            }
-            self.run('%(RUBY)s setup.rb all %(OPTIONS)s' % s, env)
+        cmd = [
+            self.ruby_executable,
+            'setup.rb',
+            'all',
+            '--prefix=%s' % self.options['location'],
+            '--no-rdoc',
+            '--no-ri',
+        ]
+
+        try:
+            self.run(cmd, env)
         finally:
             shutil.rmtree(srcdir)
             os.chdir(current_dir)
@@ -167,9 +191,39 @@ class Recipe(object):
         f = open(executable, 'w')
         f.write('\n'.join(content) + '\n\n')
         f.close()
-        os.chmod(executable, 0755)
+        os.chmod(executable, (
+            # rwx rw- rw-
+            stat.S_IRWXU |
+            stat.S_IRGRP | stat.S_IWGRP |
+            stat.S_IROTH | stat.S_IWOTH
+        ))
+
         return executable
 
+    def _install_gem(self, gemname, gem_executable, bindir):
+        cmd = [
+            gem_executable,
+            'install',
+            '--no-rdoc',
+            '--no-ri',
+            '--bindir=%s' % bindir,
+        ]
+
+        if '==' in gemname:
+            gemname, version = map(strip, gemname.split('==', 1))
+            cmd.append(gemname)
+            cmd.append('--version=%s' % version)
+        else:
+            cmd.append(gemname)
+
+        extra = self.options.get('gem-options', '')
+        extra = filter(None, map(strip, extra.splitlines()))
+
+        cmd.append('--')
+        cmd.extend(extra)
+
+        self.run(cmd, self._get_env())
+
     def get_gem_executable(self, bindir):
         gem_executable = os.path.join(bindir, 'gem')
         gem_executable = glob.glob(gem_executable + '*')
@@ -188,29 +242,13 @@ class Recipe(object):
             gem_executable = self.get_gem_executable(bindir)
 
         for gemname in self.gems:
-            extra = self.options.get('gem-options', '')
-            extra = ' '.join(filter(None, map(strip, extra.splitlines())))
-            s = {
-                'GEM': gem_executable,
-                'OPTIONS': ' '.join([
-                        '--no-rdoc',
-                        '--no-ri',
-                        '--bindir=%s' % bindir,
-                    ]),
-                'EXTRA': extra,
-            }
-            if '==' in gemname:
-                gemname, version = map(strip, gemname.split('==', 1))
-                s['GEMNAME'] = gemname
-                s['OPTIONS'] += ' --version %s' % version
-            else:
-                s['GEMNAME'] = gemname
-            self.run('%(GEM)s install %(OPTIONS)s %(GEMNAME)s -- %(EXTRA)s' % s,
-                     self._get_env())
+            self.log.info('installing ruby gem "%s"' % gemname)
+            self._install_gem(gemname, gem_executable, bindir)
 
         for executable in os.listdir(bindir):
             installed_path = self._install_executable(
-                    os.path.join(bindir, executable))
+                os.path.join(bindir, executable)
+            )
             parts.append(installed_path)
 
         return parts
